﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading;

namespace Concurrency
{
    public delegate int Compare<T>(T x, T y);

    public class ConcurrentSkipList<T> : ICollection<T> where T : IComparable<T>
    {
        class Node
        {
            internal T key;
            internal int topLayer;
            internal Node[] nexts;
            internal volatile bool marked = false;
            internal volatile bool fullyLinked = false;

            internal Node(T k, int height)
            {
                key = k;
                topLayer = height;
                nexts = new Node[topLayer + 1];
            }
        }

        readonly Node LSentinel, RSentinel;

        Random r = new Random();
        const int MaxHeight = sizeof(uint) * 8;

        uint randomSeed;
        /// <summary>
        /// Copied straight out of java.util.concurrent.ConcurrentSkipListMap, except the last line.
        /// </summary>
        /// <returns></returns>
        int randomLevel()
        {
            uint x = randomSeed;
            x ^= x << 13;
            x ^= x >> 17;
            randomSeed = x ^= x << 5;
            if ((x & 0x8001) != 0) // test highest and lowest bits
                return 0;
            int level = 1;
            while (((x >>= 1) & 1) != 0) ++level;
            return level;
        }

        bool okToDelete(Node candidate, int lFound)
        {
            return candidate.fullyLinked
                && candidate.topLayer == lFound
                && !candidate.marked;
        }

        int? findNode(T v, Node[] preds, Node[] succs)
        {
            int? lFound = null;
            Node pred = LSentinel;
            for (int layer = MaxHeight - 1; layer >= 0; layer--)
            {
                Node curr = pred.nexts[layer];
                while (curr != RSentinel && v.CompareTo(curr.key) > 0)
                {
                    pred = curr;
                    curr = pred.nexts[layer];
                }
                if (!lFound.HasValue && curr != RSentinel && v.CompareTo(curr.key) == 0)
                    lFound = layer;
                preds[layer] = pred;
                succs[layer] = curr;
            }
            return lFound;
        }

        public int count = 0;
        public int addRetry = 0;
        public int removeRetry = 0;

        public bool contains(T v)
        {
            Node[] preds = new Node[MaxHeight];
            Node[] succs = new Node[MaxHeight];
            int? lFound = findNode(v, preds, succs);
            return lFound.HasValue
                && succs[lFound.Value].fullyLinked
                && !succs[lFound.Value].marked;
        }

        public bool add(T v)
        {
            int topLayer = randomLevel();
            Node[] preds = new Node[MaxHeight];
            Node[] succs = new Node[MaxHeight];

            bool firstTime = true;
            while (true)
            {
                if (!firstTime) addRetry++; 

                int? lFound = findNode(v, preds, succs);
                if (lFound.HasValue)
                {
                    Node nodeFound = succs[lFound.Value];
                    if (!nodeFound.marked)
                    {
                        while (!nodeFound.fullyLinked) ;
                        return false;
                    }
                    continue;
                }
                int highestLocked = -1;
                try
                {
                    Node pred, succ, prevPred = null;
                    bool valid = true;
                    for (int layer = 0; valid && (layer <= topLayer); layer++)
                    {
                        pred = preds[layer];
                        succ = succs[layer];
                        if (pred != prevPred)
                        {
                            Monitor.Enter(pred);
                            highestLocked = layer;
                            prevPred = pred;
                        }
                        valid = !pred.marked && !succ.marked && pred.nexts[layer] == succ;
                    }
                    if (!valid) continue;

                    Node newNode = new Node(v, topLayer);
                    for (int layer = 0; layer <= topLayer; layer++)
                    {
                        newNode.nexts[layer] = succs[layer];
                        preds[layer].nexts[layer] = newNode;
                    }

                    newNode.fullyLinked = true;
                    Interlocked.Increment(ref count);
                    return true;
                }
                finally
                {
                    firstTime = false;
                    Node previous = null;
                    for (int i = 0; i <= highestLocked; i++)
                    {
                        if (preds[i] != previous)
                        {
                            previous = preds[i];
                            Monitor.Exit(preds[i]);
                        }
                    }
                }
            }
        }

        public bool remove(T v)
        {
            Node nodeToDelete = null;
            bool isMarked = false;
            int topLayer = -1;
            Node[] preds = new Node[MaxHeight];
            Node[] succs = new Node[MaxHeight];

            bool firstTime = true;

            while (true)
            {
                if (!firstTime) removeRetry++; 

                int? lFound = findNode(v, preds, succs);
                if (isMarked || (lFound.HasValue && okToDelete(succs[lFound.Value], lFound.Value)))
                {
                    if (!isMarked)
                    {
                        nodeToDelete = succs[lFound.Value];
                        topLayer = nodeToDelete.topLayer;
                        Monitor.Enter(nodeToDelete);
                        if (nodeToDelete.marked)
                        {
                            Monitor.Exit(nodeToDelete);
                            return false;
                        }
                        nodeToDelete.marked = true;
                        isMarked = true;
                    }
                    int highestLocked = -1;
                    try
                    {
                        Node pred, succ, prevPred = null;
                        bool valid = true;
                        for (int layer = 0; valid && (layer <= topLayer); layer++)
                        {
                            pred = preds[layer];
                            succ = succs[layer];
                            if (pred != prevPred)
                            {
                                Monitor.Enter(pred);
                                highestLocked = layer;
                                prevPred = pred;
                            }
                            valid = !pred.marked && pred.nexts[layer] == succ;
                        }
                        if (!valid) continue;

                        for (int layer = topLayer; layer >= 0; layer--)
                            preds[layer].nexts[layer] = nodeToDelete.nexts[layer];
                        Monitor.Exit(nodeToDelete);
                        Interlocked.Decrement(ref count);
                        return true;
                    }
                    finally
                    {
                        firstTime = false;
                        Node previous = null;
                        for (int i = 0; i <= highestLocked; i++)
                        {
                            if (preds[i] != previous)
                            {
                                previous = preds[i];
                                Monitor.Exit(preds[i]);
                            }
                        }
                    }
                }
                else return false;
            }
        }

        List<T> toList()
        {
            List<T> l = new List<T>(count);
            for (Node curr = LSentinel.nexts[0]; curr != RSentinel; curr = curr.nexts[0])
            {
                if (!curr.marked)
                {
                    while (!curr.fullyLinked) ;
                    l.Add(curr.key);
                }
            }
            return l;
        }

        public ConcurrentSkipList()
        {
            randomSeed = (uint)r.Next() | 0x0100;
            LSentinel = new Node(default(T), MaxHeight);
            RSentinel = new Node(default(T), MaxHeight);
            for (int layer = 0; layer < MaxHeight; layer++)
                LSentinel.nexts[layer] = RSentinel;
        }

        #region ICollection<T> Members

        void ICollection<T>.Add(T item)
        {
            add(item);
        }

        void ICollection<T>.Clear()
        {
            throw new NotImplementedException();
        }

        bool ICollection<T>.Contains(T item)
        {
            return contains(item);
        }

        void ICollection<T>.CopyTo(T[] array, int arrayIndex)
        {
            toList().CopyTo(array, arrayIndex);
        }

        int ICollection<T>.Count
        {
            get { return count; }
        }

        bool ICollection<T>.IsReadOnly
        {
            get { return false; }
        }

        bool ICollection<T>.Remove(T item)
        {
            return remove(item);
        }

        #endregion

        #region IEnumerable<T> Members

        IEnumerator<T> IEnumerable<T>.GetEnumerator()
        {
            return toList().GetEnumerator();
        }

        #endregion

        #region IEnumerable Members

        System.Collections.IEnumerator System.Collections.IEnumerable.GetEnumerator()
        {
            return toList().GetEnumerator();
        }

        #endregion
    }
}
